import pickle
import os
import tensorflow as tf
import numpy as np
from collections import defaultdict
from tqdm import tqdm


from capreolus.extractor import Extractor
from capreolus import Dependency, ConfigOption, get_logger
from capreolus.utils.common import padlist
from capreolus.utils.exceptions import MissingDocError
from capreolus.tokenizer.punkt import PunktTokenizer

logger = get_logger(__name__)


@Extractor.register
class BertPassage(Extractor):
    """
    Extracts passages from the document to be later consumed by a BERT based model.
    Does NOT use all the passages. The first passages is always used. Use the `prob` config to control the probability
    of a passage being selected
    Gotcha: In Tensorflow the train tfrecords have shape (batch_size, maxseqlen) while dev tf records have the shape
    (batch_size, num_passages, maxseqlen). This is because during inference, we want to pool over the scores of the
    passages belonging to a doc
    """

    module_name = "bertpassage"
    dependencies = [
        Dependency(
            key="index", module="index", name="anserini", default_config_overrides={"indexstops": True, "stemmer": "none"}
        ),
        Dependency(key="tokenizer", module="tokenizer", name="berttokenizer"),
    ]

    pad = 0
    pad_tok = "[PAD]"

    config_spec = [
        ConfigOption("maxseqlen", 256, "Maximum input length for BERT"),
        ConfigOption("usecache", False, "Should the extracted features be cached?"),
        ConfigOption("passagelen", 150, "Length of the extracted passage"),
        ConfigOption("stride", 100, "Stride"),
        ConfigOption("sentences", False, "Use a sentence tokenizer to form passages"),
        ConfigOption("numpassages", 16, "Number of passages per document"),
        ConfigOption(
            "prob",
            0.1,
            "The probability that a passage from the document will be used for training (the first passage is always used)",
        ),
    ]

    def load_state(self, qids, docids):
        cache_fn = self.get_state_cache_file_path(qids, docids)
        logger.debug("loading state from: %s", cache_fn)
        with open(cache_fn, "rb") as f:
            state_dict = pickle.load(f)
            self.qid2toks = state_dict["qid2toks"]
            self.docid2passages = state_dict["docid2passages"]

    def cache_state(self, qids, docids):
        os.makedirs(self.get_cache_path(), exist_ok=True)
        with open(self.get_state_cache_file_path(qids, docids), "wb") as f:
            state_dict = {"qid2toks": self.qid2toks, "docid2passages": self.docid2passages}
            pickle.dump(state_dict, f, protocol=-1)

    def get_tf_feature_description(self):
        feature_description = {
            "pos_bert_input": tf.io.FixedLenFeature([], tf.string),
            "pos_mask": tf.io.FixedLenFeature([], tf.string),
            "pos_seg": tf.io.FixedLenFeature([], tf.string),
            "neg_bert_input": tf.io.FixedLenFeature([], tf.string),
            "neg_mask": tf.io.FixedLenFeature([], tf.string),
            "neg_seg": tf.io.FixedLenFeature([], tf.string),
            "label": tf.io.FixedLenFeature([], tf.string),
        }

        return feature_description

    def create_tf_train_feature(self, sample):
        """
        Returns a set of features from a doc.
        Of the num_passages passages that are present in a document, we use only a subset of it.
        params:
        sample - A dict where each entry has the shape [batch_size, num_passages, maxseqlen]

        Returns a list of features. Each feature is a dict, and each value in the dict has the shape [batch_size, maxseqlen].
        Yes, the output shape is different to the input shape because we sample from the passages.
        """
        num_passages = self.config["numpassages"]

        def _bytes_feature(value):
            """Returns a bytes_list from a string / byte. Our features are multi-dimensional tensors."""
            if isinstance(value, type(tf.constant(0))):  # if value ist tensor
                value = value.numpy()  # get value of tensor
            return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

        posdoc, negdoc, negdoc_id = sample["pos_bert_input"], sample["neg_bert_input"], sample["negdocid"]
        posdoc_mask, posdoc_seg, negdoc_mask, negdoc_seg = (
            sample["pos_mask"],
            sample["pos_seg"],
            sample["neg_mask"],
            sample["neg_seg"],
        )
        label = sample["label"]
        features = []

        for i in range(num_passages):
            # Always use the first passage, then sample from the remaining passages
            if i > 0 and self.rng.random() > self.config["prob"]:
                continue

            bert_input_line = posdoc[i]
            bert_input_line = " ".join(self.tokenizer.bert_tokenizer.convert_ids_to_tokens(list(bert_input_line)))
            passage = bert_input_line.split("[SEP]")[-2]

            # Ignore empty passages as well
            if passage.strip() == "[PAD]":
                continue

            feature = {
                "pos_bert_input": _bytes_feature(tf.io.serialize_tensor(posdoc[i])),
                "pos_mask": _bytes_feature(tf.io.serialize_tensor(posdoc_mask[i])),
                "pos_seg": _bytes_feature(tf.io.serialize_tensor(posdoc_seg[i])),
                "neg_bert_input": _bytes_feature(tf.io.serialize_tensor(negdoc[i])),
                "neg_mask": _bytes_feature(tf.io.serialize_tensor(negdoc_mask[i])),
                "neg_seg": _bytes_feature(tf.io.serialize_tensor(negdoc_seg[i])),
                "label": _bytes_feature(tf.io.serialize_tensor(label[i])),
            }
            features.append(feature)

        return features

    def create_tf_dev_feature(self, sample):
        """
        Unlike the train feature, the dev set uses all passages. Both the input and the output are dicts with the shape
        [batch_size, num_passages, maxseqlen]
        """
        posdoc, negdoc, negdoc_id = sample["pos_bert_input"], sample["neg_bert_input"], sample["negdocid"]
        posdoc_mask, posdoc_seg, negdoc_mask, negdoc_seg = (
            sample["pos_mask"],
            sample["pos_seg"],
            sample["neg_mask"],
            sample["neg_seg"],
        )
        label = sample["label"]

        def _bytes_feature(value):
            """Returns a bytes_list from a string / byte."""
            if isinstance(value, type(tf.constant(0))):  # if value ist tensor
                value = value.numpy()  # get value of tensor
            return tf.train.Feature(bytes_list=tf.train.BytesList(value=[value]))

        feature = {
            "pos_bert_input": _bytes_feature(tf.io.serialize_tensor(posdoc)),
            "pos_mask": _bytes_feature(tf.io.serialize_tensor(posdoc_mask)),
            "pos_seg": _bytes_feature(tf.io.serialize_tensor(posdoc_seg)),
            "neg_bert_input": _bytes_feature(tf.io.serialize_tensor(negdoc)),
            "neg_mask": _bytes_feature(tf.io.serialize_tensor(negdoc_mask)),
            "neg_seg": _bytes_feature(tf.io.serialize_tensor(negdoc_seg)),
            "label": _bytes_feature(tf.io.serialize_tensor(label)),
        }

        return [feature]

    def parse_tf_train_example(self, example_proto):
        feature_description = self.get_tf_feature_description()
        parsed_example = tf.io.parse_example(example_proto, feature_description)

        def parse_tensor_as_int(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.int64)
            parsed_tensor.set_shape([self.config["maxseqlen"]])

            return parsed_tensor

        def parse_label_tensor(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.float32)
            parsed_tensor.set_shape([2])

            return parsed_tensor

        pos_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["pos_bert_input"], dtype=tf.int64)
        pos_mask = tf.map_fn(parse_tensor_as_int, parsed_example["pos_mask"], dtype=tf.int64)
        pos_seg = tf.map_fn(parse_tensor_as_int, parsed_example["pos_seg"], dtype=tf.int64)
        neg_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["neg_bert_input"], dtype=tf.int64)
        neg_mask = tf.map_fn(parse_tensor_as_int, parsed_example["neg_mask"], dtype=tf.int64)
        neg_seg = tf.map_fn(parse_tensor_as_int, parsed_example["neg_seg"], dtype=tf.int64)
        label = tf.map_fn(parse_label_tensor, parsed_example["label"], dtype=tf.float32)

        return (pos_bert_input, pos_mask, pos_seg, neg_bert_input, neg_mask, neg_seg), label

    def parse_tf_dev_example(self, example_proto):
        feature_description = self.get_tf_feature_description()
        parsed_example = tf.io.parse_example(example_proto, feature_description)

        def parse_tensor_as_int(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.int64)
            parsed_tensor.set_shape([self.config["numpassages"], self.config["maxseqlen"]])

            return parsed_tensor

        def parse_label_tensor(x):
            parsed_tensor = tf.io.parse_tensor(x, tf.float32)
            parsed_tensor.set_shape([self.config["numpassages"], 2])

            return parsed_tensor

        pos_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["pos_bert_input"], dtype=tf.int64)
        pos_mask = tf.map_fn(parse_tensor_as_int, parsed_example["pos_mask"], dtype=tf.int64)
        pos_seg = tf.map_fn(parse_tensor_as_int, parsed_example["pos_seg"], dtype=tf.int64)
        neg_bert_input = tf.map_fn(parse_tensor_as_int, parsed_example["neg_bert_input"], dtype=tf.int64)
        neg_mask = tf.map_fn(parse_tensor_as_int, parsed_example["neg_mask"], dtype=tf.int64)
        neg_seg = tf.map_fn(parse_tensor_as_int, parsed_example["neg_seg"], dtype=tf.int64)
        label = tf.map_fn(parse_label_tensor, parsed_example["label"], dtype=tf.float32)

        return (pos_bert_input, pos_mask, pos_seg, neg_bert_input, neg_mask, neg_seg), label

    def get_passages_for_doc(self, doc):
        """
        Extract passages from the doc.
        If there are too many passages, keep the first and the last one and sample from the rest.
        If there are not enough packages, pad.
        """
        tokenize = self.tokenizer.tokenize
        numpassages = self.config["numpassages"]
        passages = []

        for i in range(0, len(doc), self.config["stride"]):
            if i >= len(doc):
                assert len(passages) > 0, f"no passage can be built from empty document {doc}"
                break
            else:
                passage = doc[i : i + self.config["passagelen"]]

            passages.append(tokenize(" ".join(passage)))

        n_actual_passages = len(passages)
        # If we have a more passages than required, keep the first and last, and sample from the rest
        if n_actual_passages > numpassages:
            if numpassages > 1:
                passages = [passages[0]] + list(self.rng.choice(passages[1:-1], numpassages - 2, replace=False)) + [passages[-1]]
            else:
                passages = [passages[0]]
        else:
            # Pad until we have the required number of passages
            for _ in range(numpassages - n_actual_passages):
                passages.append(["[PAD]"])

        assert len(passages) == self.config["numpassages"]
        return passages

    # from https://github.com/castorini/birch/blob/2dd0401ebb388a1c96f8f3357a064164a5db3f0e/src/utils/doc_utils.py#L73
    def _chunk_sent(self, sent, max_len):
        words = self.tokenizer.tokenize(sent)

        if len(words) <= max_len:
            return [words]

        chunked_sents = []
        size = int(len(words) / max_len)
        for i in range(0, size):
            seq = words[i * max_len : (i + 1) * max_len]
            chunked_sents.append(seq)
        return chunked_sents

    def _build_passages_from_sentences(self, docids):
        punkt = PunktTokenizer()

        for docid in tqdm(docids, "extract passages"):
            passages = []
            numpassages = self.config["numpassages"]
            for sentence in punkt.tokenize(self.index.get_doc(docid)):
                if len(passages) >= numpassages:
                    break

                passages.extend(self._chunk_sent(sentence, self.config["passagelen"]))

            if numpassages != 0:
                passages = passages[:numpassages]

                n_actual_passages = len(passages)
                for _ in range(numpassages - n_actual_passages):
                    # randomly use one of previous passages when the document is exhausted
                    # idx = random.randint(0, n_actual_passages - 1)
                    # passages.append(passages[idx])

                    # append empty passages
                    passages.append([""])

                assert len(passages) == self.config["numpassages"]

            self.docid2passages[docid] = sorted(passages, key=len)

    def _build_vocab(self, qids, docids, topics):
        if self.is_state_cached(qids, docids) and self.config["usecache"]:
            self.load_state(qids, docids)
            logger.info("Vocabulary loaded from cache")
        elif self.config["sentences"]:
            self.docid2passages = {}
            self._build_passages_from_sentences(docids)
            self.qid2toks = {qid: self.tokenizer.tokenize(topics[qid]) for qid in tqdm(qids, desc="querytoks")}
            self.cache_state(qids, docids)
        else:
            logger.info("Building bertpassage vocabulary")
            self.docid2passages = {}

            for docid in tqdm(docids, "extract passages"):
                # Naive tokenization based on white space
                doc = self.index.get_doc(docid).split()
                passages = self.get_passages_for_doc(doc)
                self.docid2passages[docid] = passages

            self.qid2toks = {qid: self.tokenizer.tokenize(topics[qid]) for qid in tqdm(qids, desc="querytoks")}
            self.cache_state(qids, docids)

    def exist(self):
        return hasattr(self, "docid2passages") and len(self.docid2passages)

    def preprocess(self, qids, docids, topics):
        if self.exist():
            return

        self.index.create_index()
        self.qid2toks = defaultdict(list)
        self.docid2passages = None

        self._build_vocab(qids, docids, topics)

    def id2vec(self, qid, posid, negid=None, label=None):
        """
        See parent class for docstring
        """
        assert label is not None

        tokenizer = self.tokenizer
        maxseqlen = self.config["maxseqlen"]

        query_toks = self.qid2toks[qid]
        pos_bert_inputs = []
        pos_bert_masks = []
        pos_bert_segs = []

        # N.B: The passages in self.docid2passages are not bert tokenized
        pos_passages = self.docid2passages[posid]
        for tokenized_passage in pos_passages:
            input_line = ["[CLS]"] + query_toks + ["[SEP]"] + tokenized_passage + ["[SEP]"]
            if len(input_line) > maxseqlen:
                input_line = input_line[:maxseqlen]
                input_line[-1] = "[SEP]"

            padded_input_line = padlist(input_line, padlen=self.config["maxseqlen"], pad_token=self.pad_tok)
            pos_bert_masks.append([1] * len(input_line) + [0] * (len(padded_input_line) - len(input_line)))
            pos_bert_segs.append([0] * (len(query_toks) + 2) + [1] * (len(padded_input_line) - len(query_toks) - 2))
            pos_bert_inputs.append(tokenizer.convert_tokens_to_ids(padded_input_line))

        # TODO: Rename the posdoc key in the below dict to 'pos_bert_input'
        data = {
            "qid": qid,
            "posdocid": posid,
            "pos_bert_input": np.array(pos_bert_inputs, dtype=np.long),
            "pos_mask": np.array(pos_bert_masks, dtype=np.long),
            "pos_seg": np.array(pos_bert_segs, dtype=np.long),
            "negdocid": "",
            "neg_bert_input": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "neg_mask": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "neg_seg": np.zeros((self.config["numpassages"], self.config["maxseqlen"]), dtype=np.long),
            "label": np.repeat(np.array([label], dtype=np.float32), self.config["numpassages"], 0),
        }

        if negid:
            neg_bert_inputs = []
            neg_bert_masks = []
            neg_bert_segs = []
            neg_passages = self.docid2passages[negid]
            for tokenized_passage in neg_passages:
                input_line = ["[CLS]"] + query_toks + ["[SEP]"] + tokenized_passage + ["[SEP]"]
                if len(input_line) > maxseqlen:
                    input_line = input_line[:maxseqlen]
                    input_line[-1] = "[SEP]"

                padded_input_line = padlist(input_line, padlen=self.config["maxseqlen"], pad_token=self.pad_tok)
                neg_bert_masks.append([1] * len(input_line) + [0] * (len(padded_input_line) - len(input_line)))
                neg_bert_segs.append([0] * (len(query_toks) + 2) + [1] * (len(padded_input_line) - len(query_toks) - 2))
                neg_bert_inputs.append(tokenizer.convert_tokens_to_ids(padded_input_line))

            if not neg_bert_inputs:
                raise MissingDocError(qid, negid)

            data["negdocid"] = negid
            data["neg_bert_input"] = np.array(neg_bert_inputs, dtype=np.long)
            data["neg_mask"] = np.array(neg_bert_masks, dtype=np.long)
            data["neg_seg"] = np.array(neg_bert_segs, dtype=np.long)

        return data



from transformers import BertTokenizerFast

from capreolus import ConfigOption

from . import Tokenizer


@Tokenizer.register
class BertTokenizer(Tokenizer):
    module_name = "berttokenizer"
    config_spec = [ConfigOption("pretrained", "bert-base-uncased", "pretrained model to load vocab from")]

    def build(self):
        self.bert_tokenizer = BertTokenizerFast.from_pretrained(self.config["pretrained"])

    def convert_tokens_to_ids(self, tokens):
        return self.bert_tokenizer.convert_tokens_to_ids(tokens)

    def tokenize(self, sentences):
        if not sentences or len(sentences) == 0:  # either "" or []
            return []

        if isinstance(sentences, str):
            return self.bert_tokenizer.tokenize(sentences)

        return [self.bert_tokenizer.tokenize(s) for s in sentences]



from pathlib import Path

import pytest
from profane import constants

from capreolus.index import AnseriniIndex


@pytest.fixture(scope="function")
def tmpdir_as_cache(tmpdir, monkeypatch):
    constants._d["CACHE_BASE_PATH"] = Path(tmpdir)


@pytest.fixture(scope="function")
def dummy_index(tmpdir_as_cache):
    index = AnseriniIndex({"name": "anserini", "indexstops": False, "stemmer": "porter", "collection": {"name": "dummy"}})
    index.create_index()
    return index



import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install
import os


class PostDevelopCommand(develop):
    """Post-installation for development mode."""

    def run(self):
        develop.run(self)


class PostInstallCommand(install):
    """Post-installation for installation mode."""

    def run(self):
        install.run(self)


with open("README.md", "r") as fh:
    long_description = fh.read()

# from https://packaging.python.org/guides/single-sourcing-package-version/
def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), "rt") as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]

    raise RuntimeError("Unable to find version string.")


setuptools.setup(
    name="capreolus",
    version=get_version("capreolus/__init__.py"),
    author="Andrew Yates",
    author_email="initial-then-last@mpi-inf.mpg.de",
    description="A toolkit for end-to-end neural ad hoc retrieval",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://capreolus.ai",
    packages=setuptools.find_packages(),
    install_requires=[
        "torch==1.6",
        "torchvision==0.7",
        "cython",
        "pre-commit",
        "profane>=0.2.0",
        "SQLAlchemy",
        "sqlalchemy-utils",
        "psycopg2-binary",
        "matplotlib",
        "colorlog==4.0.2",
        "pytrec_eval>=0.5",
        "pytest",
        "pyjnius==1.2.1",
        "pymagnitude==0.1.143",
        "h5py",
        "nltk==3.4.5",
        "pytest-mock",
        "mock",
        "pyserini==0.9.3.0",
        "numpy",
        "scipy",
        "google-api-python-client",
        "oauth2client",
        "tensorflow==2.3.1",
        "transformers==3.1.0",
        "tensorflow-ranking",
        "Pillow",
        "beautifulsoup4",
        "lxml",
        "scispacy",
        "spacy",
        "pandas",
    ],
    classifiers=["Programming Language :: Python :: 3", "Operating System :: OS Independent"],
    python_requires=">=3.6",
    cmdclass={"develop": PostDevelopCommand, "install": PostInstallCommand},
    include_package_data=True,
    scripts=["scripts/capreolus"],
)



import tensorflow as tf

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import RbfKernelBankTF, similarity_matrix_tf


class TFKNRM_Class(tf.keras.layers.Layer):
    def __init__(self, extractor, config, **kwargs):
        super(TFKNRM_Class, self).__init__(**kwargs)
        self.config = config
        self.extractor = extractor
        mus = [-0.9, -0.7, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0]
        sigmas = [0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.1, 0.001]
        self.embedding = tf.keras.layers.Embedding(
            len(self.extractor.stoi), self.extractor.embeddings.shape[1], weights=[self.extractor.embeddings], trainable=False
        )
        self.kernels = RbfKernelBankTF(mus, sigmas, dim=1, requires_grad=config["gradkernels"])
        self.combine = tf.keras.layers.Dense(1, input_shape=(self.kernels.count(),))

    def get_score(self, doc_tok, query_tok, query_idf):
        query = self.embedding(query_tok)
        doc = self.embedding(doc_tok)
        batch_size, qlen, doclen = tf.shape(query)[0], tf.shape(query)[1], tf.shape(doc)[1]

        simmat = similarity_matrix_tf(query, doc, query_tok, doc_tok, self.extractor.pad)

        k = self.kernels(simmat)
        doc_k = tf.reduce_sum(k, axis=3)  # sum over document
        reshaped_simmat = tf.broadcast_to(
            tf.reshape(simmat, (batch_size, 1, qlen, doclen)), (batch_size, self.kernels.count(), qlen, doclen)
        )
        mask = tf.reduce_sum(reshaped_simmat, axis=3) != 0.0
        log_k = tf.where(mask, tf.math.log(doc_k + 1e-6), tf.cast(mask, tf.float32))
        query_k = tf.reduce_sum(log_k, axis=2)
        scores = self.combine(query_k)

        return tf.reshape(scores, [batch_size])

    def call(self, x, **kwargs):
        doc, query, query_idf = x[0], x[1], x[2]
        score = self.get_score(doc, query, query_idf)

        return score

    def predict_step(self, data):
        return self.score(data)

    def score(self, x, **kwargs):
        posdoc, negdoc, query, query_idf = x

        return self.call((posdoc, query, query_idf))

    def score_pair(self, x, **kwargs):
        posdoc, negdoc, query, query_idf = x
        pos_score = self.call((posdoc, query, query_idf))
        neg_score = self.call((negdoc, query, query_idf))

        return pos_score, neg_score


@Reranker.register
class TFKNRM(Reranker):
    """TensorFlow implementation of KNRM.

    Chenyan Xiong, Zhuyun Dai, Jamie Callan, Zhiyuan Liu, and Russell Power. 2017. End-to-End Neural Ad-hoc Ranking with Kernel Pooling. In SIGIR'17.
    """

    module_name = "TFKNRM"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="tensorflow"),
    ]
    config_spec = [
        ConfigOption("gradkernels", True, "backprop through mus and sigmas"),
        ConfigOption("finetune", False, "fine tune the embedding layer"),  # TODO check save when True
    ]

    def build_model(self):
        self.model = TFKNRM_Class(self.extractor, self.config)

        return self.model



import torch
from torch import nn
from torch.autograd import Variable

from capreolus import Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import create_emb_layer
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name

device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")


class POSITDRMM_basic(nn.Module):
    def __init__(self, extractor, pipeline_config):
        super(POSITDRMM_basic, self).__init__()
        #        self.embedding_dim=embedding_dim
        weights_matrix = extractor.embeddings
        self.embedding_dim = weights_matrix.shape[1]
        self.lstm_hidden_dim = weights_matrix.shape[1]
        self.batch_size = pipeline_config["batch"]
        self.QUERY_LENGTH = pipeline_config["maxqlen"]
        #        self.lstm_hidden_dim = lstm_hidden_dim
        self.lstm_num_layers = 2
        self.encoding_layer = nn.LSTM(
            input_size=self.embedding_dim,
            hidden_size=self.embedding_dim,
            num_layers=self.lstm_num_layers,
            bidirectional=True,
            dropout=0.3,
        )
        self.pad_token = extractor.pad
        self.embedding = create_emb_layer(weights_matrix, non_trainable=True)
        self.m = nn.Dropout(p=0.2)
        self.Q1 = nn.Linear(6, 1, bias=True)
        self.Wg = nn.Linear(5, 1)
        self.activation = nn.LeakyReLU()
        self.hidden = self.init_hidden()

    def init_hidden(self):

        return (
            Variable(torch.zeros(4, self.batch_size, self.lstm_hidden_dim).to(device)),
            Variable(torch.zeros(4, self.batch_size, self.lstm_hidden_dim).to(device)),
        )

    def forward(self, sentence, query_sentence, query_idf, extra):
        dtype = torch.FloatTensor
        x = self.embedding(sentence)
        query_x = self.embedding(query_sentence)
        x1 = x.norm(dim=2)[:, :, None] + 1e-7
        query_x1 = query_x.norm(dim=2)[:, :, None] + 1e-7
        x_norm = x / x1
        query_x_norm = query_x / query_x1
        M_cos = torch.matmul(query_x_norm, torch.transpose(x_norm, 1, 2))
        BAT = query_sentence.shape[0]
        A = query_sentence.shape[1]
        B = sentence.shape[1]
        nul = torch.zeros_like(M_cos)
        one = torch.ones_like(M_cos)
        XOR_matrix = torch.where(
            query_sentence.reshape(BAT, A, 1).expand(BAT, A, B) == sentence.reshape(BAT, 1, B).expand(BAT, A, B), one, nul
        )
        XOR_matrix = torch.where(query_sentence.reshape(BAT, A, 1).expand(BAT, A, B) == self.pad_token, nul, XOR_matrix)
        query_x = torch.transpose(query_x, 0, 1)
        (p1, self.hidden) = self.encoding_layer(self.m(query_x))
        p1_forward = p1[:, :, : (self.lstm_hidden_dim)]
        p1_backward = p1[:, :, (self.lstm_hidden_dim) :]
        query_context = torch.cat([p1_forward + query_x, p1_backward + query_x], dim=2)
        x = torch.transpose(x, 0, 1)
        (p1, self.hidden) = self.encoding_layer(self.m(x))
        p1_forward = p1[:, :, : (self.lstm_hidden_dim)]
        p1_backward = p1[:, :, (self.lstm_hidden_dim) :]
        doc_context = torch.cat([p1_forward + x, p1_backward + x], dim=2)
        query_context = torch.transpose(query_context, 0, 1)
        doc_context = torch.transpose(doc_context, 0, 1)
        doc_context_norm = doc_context / doc_context.norm(dim=2)[:, :, None]
        query_con_norm = query_context / query_context.norm(dim=2)[:, :, None]
        M_cos_context = torch.matmul(query_con_norm, torch.transpose(doc_context_norm, 1, 2))
        M_max = torch.randn(self.batch_size, self.QUERY_LENGTH, 6).type(dtype).to(sentence.device)
        M_max[:, :, 0], _ = torch.max(M_cos, 2)
        i, _ = torch.topk(M_cos, 5, dim=2)
        M_max[:, :, 1] = torch.sum(i, dim=2) / 5
        M_max[:, :, 2], _ = torch.max(M_cos_context, 2)
        i, _ = torch.topk(M_cos_context, 5, dim=2)
        M_max[:, :, 3] = torch.sum(i, dim=2) / 5
        M_max[:, :, 4], _ = torch.max(XOR_matrix, 2)
        i, _ = torch.topk(XOR_matrix, 5, dim=2)
        M_max[:, :, 5] = torch.sum(i, dim=2) / 5
        M_res = self.activation(self.Q1(M_max))
        M_res = M_res.view(self.batch_size, self.QUERY_LENGTH)
        mmm = nn.Softmax(dim=1)

        r = mmm(query_idf.float())
        s = r * M_res
        ans = torch.sum(s, dim=1).view(-1, 1)
        extra_ans = torch.cat([ans, extra], dim=1)
        true_ans = self.Wg(extra_ans)
        return true_ans.view(-1)


class POSITDRMM_class(nn.Module):
    def __init__(self, extractor, config):
        super(POSITDRMM_class, self).__init__()
        self.posit1 = POSITDRMM_basic(extractor, config)
        self.p = config

    def forward(self, query_sentence, query_idf, pos_sentence, neg_sentence, posdoc_extra, negdoc_extra):
        self.posit1.hidden = self.posit1.init_hidden()
        pos_tag_scores = self.posit1(pos_sentence, query_sentence, query_idf, posdoc_extra)
        self.posit1.hidden = self.posit1.init_hidden()
        neg_tag_scores = self.posit1(neg_sentence, query_sentence, query_idf, negdoc_extra)
        return [pos_tag_scores, neg_tag_scores]

    def test_forward(self, query_sentence, query_idf, pos_sentence, extras):
        self.posit1.hidden = self.posit1.init_hidden()
        pos_tag_scores = self.posit1(pos_sentence, query_sentence, query_idf, extras)
        #        neg_tag_scores = self.posit1(neg_sentence, query_sentence, query_idf)
        #        self.posit1.hidden = self.posit1.init_hidden()
        return pos_tag_scores


dtype = torch.FloatTensor


@Reranker.register
class POSITDRMM(Reranker):
    """Ryan McDonald, George Brokos, and Ion Androutsopoulos. 2018. Deep Relevance Ranking Using Enhanced Document-Query Interactions. In EMNLP'18."""

    module_name = "POSITDRMM"
    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            config = dict(self.config)
            config["batch"] = self.trainer.config["batch"]
            config.update(self.extractor.config)
            self.model = POSITDRMM_class(self.extractor, config)

        return self.model

    def score(self, data):
        query_idf = data["query_idf"]
        query_sentence = data["query"]
        pos_sentence, neg_sentence = data["posdoc"], data["negdoc"]
        pos_exact_matches, pos_exact_match_idf, pos_bigram_matches = self.get_exact_match_stats(
            query_idf, query_sentence, pos_sentence
        )
        neg_exact_matches, neg_exact_match_idf, neg_bigram_matches = self.get_exact_match_stats(
            query_idf, query_sentence, neg_sentence
        )
        pos_bm25 = self.get_searcher_scores(data["qid"], data["posdocid"])
        neg_bm25 = self.get_searcher_scores(data["qid"], data["negdocid"])
        posdoc_extra = torch.cat((pos_exact_matches, pos_exact_match_idf, pos_bigram_matches, pos_bm25), dim=1).to(device)
        negdoc_extra = torch.cat((neg_exact_matches, neg_exact_match_idf, neg_bigram_matches, neg_bm25), dim=1).to(device)
        return self.model(query_sentence, query_idf, pos_sentence, neg_sentence, posdoc_extra, negdoc_extra)

    def test(self, data):
        query_sentence = data["query"]
        query_idf = data["query_idf"]
        pos_sentence = data["posdoc"]
        qids = data["qid"]
        posdoc_ids = data["posdocid"]

        pos_exact_matches, pos_exact_match_idf, pos_bigram_matches = self.get_exact_match_stats(
            query_idf, query_sentence, pos_sentence
        )
        pos_bm25 = self.get_searcher_scores(qids, posdoc_ids)
        posdoc_extra = torch.cat((pos_exact_matches, pos_exact_match_idf, pos_bigram_matches, pos_bm25), dim=1).to(device)

        return self.model.test_forward(query_sentence, query_idf, pos_sentence, posdoc_extra)

    def zero_grad(self, *args, **kwargs):
        self.model.zero_grad(*args, **kwargs)

    def get_searcher_scores(self, qids, doc_ids):
        scores = torch.zeros((len(doc_ids)))
        for i, doc_id in enumerate(doc_ids):
            # The searcher_scores attribute is set from RerankTask.train()
            # TODO: Remove this temporary hack and figure out a way to pass these kind of features cleanly
            scores[i] = self.searcher_scores[qids[i]][doc_id]

        return scores.reshape(len(doc_ids), 1)

    def clean(self, text):
        """
        Remove pad tokens from the text
        """
        return text[text != self.extractor.pad]

    def get_bigrams(self, text):
        # text_batch has shape (batch_size, length)
        return torch.stack([text[i : i + 2] for i in range(len(text) - 1)])

    def get_bigram_match_count(self, query, doc):
        bigram_matches = 0
        query_bigrams = self.get_bigrams(query)
        doc_bigrams = self.get_bigrams(doc)

        for q_bigram in query_bigrams:
            bigram_matches += len(torch.all((doc_bigrams == q_bigram), dim=1).nonzero())
        return bigram_matches / len(doc_bigrams)

    def get_exact_match_count(self, query, single_doc, query_idf):
        exact_match_count = 0
        exact_match_idf = 0
        for i, q_word in enumerate(query):
            curr_count = len(single_doc[single_doc == q_word])
            exact_match_count += curr_count
            exact_match_idf += curr_count * query_idf[i]

        return exact_match_count / len(single_doc), exact_match_idf / len(single_doc)

    def get_exact_match_stats(self, query_idf_batch, query_batch, doc_batch):
        """
        The extra 4 features that must be combined to the relevant score.
        See https://www.aclweb.org/anthology/D18-1211.pdf section 4.1 beginning.
        """
        batch_size = doc_batch.shape[0]
        exact_matches = torch.zeros(batch_size)
        exact_match_idf = torch.zeros(batch_size)
        bigram_matches = torch.zeros(batch_size)

        for batch in range(batch_size):
            # The query is deliberately not cleaned, because the `query_idf` array we get is padded
            query = query_batch[batch]
            query_idf = query_idf_batch[batch]
            curr_doc = doc_batch[batch]

            # Remove pad tokens from the doc so that pad-token in query and pad-token in doc does not result in an exact
            # match count
            cleaned_doc = self.clean(curr_doc)

            curr_exact_matches, curr_exact_matches_idf = self.get_exact_match_count(query, cleaned_doc, query_idf)
            exact_matches[batch] = curr_exact_matches
            exact_match_idf[batch] = curr_exact_matches_idf
            bigram_matches[batch] = self.get_bigram_match_count(query, cleaned_doc)

        return exact_matches.reshape(batch_size, 1), exact_match_idf.reshape(batch_size, 1), bigram_matches.reshape(batch_size, 1)



import os

from capreolus import ModuleBase, get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class Trainer(ModuleBase):
    """Base class for Trainer modules. The purpose of a Trainer is to train a :class:`~capreolus.reranker.Reranker` module and use it to make predictions. Capreolus provides two trainers: :class:`~capreolus.trainer.pytorch.PytorchTrainer` and :class:`~capreolus.trainer.tensorflow.TensorFlowTrainer`

    Modules should provide:
        - a ``train`` method that trains a reranker on training and dev (validation) data
        - a ``predict`` method that uses a reranker to make predictions on data
    """

    module_type = "trainer"
    requires_random_seed = True

    def get_paths_for_early_stopping(self, train_output_path, dev_output_path):
        os.makedirs(dev_output_path, exist_ok=True)
        dev_best_weight_fn = train_output_path / "dev.best"
        weights_output_path = train_output_path / "weights"
        info_output_path = train_output_path / "info"
        os.makedirs(weights_output_path, exist_ok=True)
        os.makedirs(info_output_path, exist_ok=True)

        loss_fn = info_output_path / "loss.txt"
        # metrics_fn = dev_output_path / "metrics.json"

        return dev_best_weight_fn, weights_output_path, info_output_path, loss_fn


from profane import import_all_modules

from .pytorch import PytorchTrainer
from .tensorflow import TensorflowTrainer

import_all_modules(__file__, __package__)



import math

import torch
from torch import nn
from torch.nn import TransformerEncoder, TransformerEncoderLayer

from capreolus import ConfigOption, Dependency
from capreolus.reranker import Reranker
from capreolus.reranker.common import StackedSimilarityMatrix, create_emb_layer
from capreolus.utils.loginit import get_logger

logger = get_logger(__name__)  # pylint: disable=invalid-name


class PositionalEncoding(nn.Module):
    def __init__(self, d_model, dropout=0.1, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[: x.size(0), :]
        return self.dropout(x)


class TK_class(nn.Module):
    """
    Adapted from https://github.com/sebastian-hofstaetter/transformer-kernel-ranking/blob/master/matchmaker/models/tk.py
    TK is a neural IR model - a fusion between transformer contextualization & kernel-based scoring
    -> uses 1 transformer block to contextualize embeddings
    -> soft-histogram kernels to score interactions
    """

    def __init__(self, extractor, config):
        super(TK_class, self).__init__()

        self.embeddim = extractor.embeddings.shape[1]
        self.p = config
        self.mus = torch.tensor([-0.9, -0.7, -0.5, -0.3, -0.1, 0.1, 0.3, 0.5, 0.7, 0.9, 1.0], dtype=torch.float)
        self.mu_matrix = self.get_mu_matrix(extractor)
        self.sigma = torch.tensor(0.1, requires_grad=False)

        dropout = 0
        non_trainable = not self.p["finetune"]
        self.embedding = create_emb_layer(extractor.embeddings, non_trainable=non_trainable)
        self.cosine_module = StackedSimilarityMatrix(padding=extractor.pad)

        self.position_encoder = PositionalEncoding(self.embeddim)
        self.mixer = nn.Parameter(torch.full([1, 1, 1], 0.9, dtype=torch.float32, requires_grad=True))
        encoder_layers = TransformerEncoderLayer(self.embeddim, config["numattheads"], config["ffdim"], dropout)
        self.transformer_encoder = TransformerEncoder(encoder_layers, config["numlayers"])

        self.s_log_fcc = nn.Linear(len(self.mus), 1, bias=False)
        self.s_len_fcc = nn.Linear(len(self.mus), 1, bias=False)
        self.comb_fcc = nn.Linear(2, 1, bias=False)

        # init with small weights, otherwise the dense output is way to high for the tanh -> resulting in loss == 1 all the time
        torch.nn.init.uniform_(self.s_log_fcc.weight, -0.014, 0.014)  # inits taken from matchzoo
        torch.nn.init.uniform_(self.s_len_fcc.weight, -0.014, 0.014)  # inits taken from matchzoo

        # init with small weights, otherwise the dense output is way to high for the tanh -> resulting in loss == 1 all the time
        torch.nn.init.uniform_(self.comb_fcc.weight, -0.014, 0.014)  # inits taken from matchzoo

    def get_mu_matrix(self, extractor):
        """
        Returns a matrix of mu values that can be directly subtracted from the cosine matrix.
        This is the matrix mu in equation 5 in the paper (https://arxiv.org/pdf/2002.01854.pdf)
        """
        qlen = extractor.config["maxqlen"]
        doclen = extractor.config["maxdoclen"]

        mu_matrix = torch.zeros(len(self.mus), qlen, doclen, requires_grad=False)

        for i, mu in enumerate(self.mus):
            mu_matrix[i] = torch.full((qlen, doclen), mu)

        return mu_matrix

    def get_mask(self, embedding):
        """
        Gets a mask of shape (seq_len, seq_len). This is an additive mask, hence masked elements should be -inf
        """
        batch_size = embedding.shape[0]
        seq_len = embedding.shape[1]
        # Get a normal mask of shape (batch_size, seq_len). Entry would be 0 if a seq element should be masked
        mask = ((embedding != torch.zeros(self.embeddim).to(embedding.device)).to(dtype=embedding.dtype).sum(-1) != 0).to(
            dtype=embedding.dtype
        )

        # The square attention mask
        encoder_mask = torch.zeros(batch_size, seq_len, seq_len).to(embedding.device)
        # Set -inf on all rows corresponding to a pad token
        encoder_mask[mask == 0] = float("-inf")
        # Set -inf on all columns corresponding to a pad token (the tricky bit)
        col_mask = mask.reshape(batch_size, 1, seq_len).expand(batch_size, seq_len, seq_len)
        encoder_mask[col_mask == 0] = float("-inf")

        return torch.cat([encoder_mask] * self.p["numattheads"])

    def get_embedding(self, toks):
        """
        Overrides KNRM_Class's get_embedding to return contextualized word embeddings
        """
        embedding = self.embedding(toks)

        # Transformer layers expect input in shape (L, N, E), where L is sequence len, N is batch, E is embed dims
        reshaped_embedding = embedding.permute(1, 0, 2)
        position_encoded_embedding = self.position_encoder(reshaped_embedding)
        # TODO: Mask should be additive
        mask = self.get_mask(embedding) if self.p["usemask"] else None
        contextual_embedding = self.transformer_encoder(position_encoded_embedding, mask).permute(1, 0, 2)
        if self.p["usemixer"]:
            return self.mixer * embedding + (1 - self.mixer) * contextual_embedding
        else:
            return self.p["alpha"] * embedding + (1 - self.p["alpha"]) * contextual_embedding

    def forward(self, doctoks, querytoks, query_idf):
        batches = doctoks.shape[0]
        qlen = querytoks.shape[1]
        doclen = doctoks.shape[1]
        doc = self.get_embedding(doctoks)
        device = doc.device
        query = self.get_embedding(querytoks)
        # cosine_matrix = self.cosine_module.forward(query, doc)
        cosine_matrix = self.cosine_module.forward(query, doc, querytoks, doctoks)
        # cosine_matrix = cosine_matrix.reshape(batches, 1, qlen, doclen)
        cosine_matrix = cosine_matrix.expand(batches, len(self.mus), qlen, doclen)
        kernel_matrix = torch.exp(-torch.pow(cosine_matrix - self.mu_matrix.to(device), 2)) / (2 * torch.pow(self.sigma, 2))
        condensed_kernel_matrix = kernel_matrix.sum(3)
        s_log_k = torch.log2(condensed_kernel_matrix).sum(2)
        s_len_k = condensed_kernel_matrix.sum(2) / doclen

        s_log = self.s_log_fcc(s_log_k)
        s_len = self.s_len_fcc(s_len_k)
        score = self.comb_fcc(torch.cat([s_log, s_len], dim=1))
        return score


@Reranker.register
class TK(Reranker):
    """Sebastian Hofstätter, Markus Zlabinger, and Allan Hanbury. 2019. TU Wien @ TREC Deep Learning '19 -- Simple Contextualization for Re-ranking. In TREC '19."""

    module_name = "TK"

    dependencies = [
        Dependency(key="extractor", module="extractor", name="slowembedtext"),
        Dependency(key="trainer", module="trainer", name="pytorch"),
    ]
    config_spec = [
        ConfigOption("gradkernels", True, "backprop through mus and sigmas"),
        ConfigOption("scoretanh", False, "use a tanh on the prediction as in paper (True) or do not use a nonlinearity (False)"),
        ConfigOption("singlefc", True, "use single fully connected layer as in paper (True) or 2 fully connected layers (False)"),
        ConfigOption("projdim", 32),
        ConfigOption("ffdim", 100),
        ConfigOption("numlayers", 2),
        ConfigOption("numattheads", 10),
        ConfigOption("alpha", 0.5),
        ConfigOption("usemask", False),
        ConfigOption("usemixer", False),
        ConfigOption("finetune", False, "fine tune the embedding layer"),  # TODO check save when True
    ]

    def build_model(self):
        if not hasattr(self, "model"):
            self.model = TK_class(self.extractor, self.config)
        return self.model

    def score(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence, neg_sentence = d["posdoc"], d["negdoc"]
        return [
            self.model(pos_sentence, query_sentence, query_idf).view(-1),
            self.model(neg_sentence, query_sentence, query_idf).view(-1),
        ]

    def test(self, d):
        query_idf = d["query_idf"]
        query_sentence = d["query"]
        pos_sentence = d["posdoc"]
        return self.model(pos_sentence, query_sentence, query_idf).view(-1)



import pytest

from capreolus import module_registry
from capreolus.collection import DummyCollection
from capreolus.index import Index
from capreolus.tests.common_fixtures import dummy_index, tmpdir_as_cache

indexs = set(module_registry.get_module_names("index"))


@pytest.mark.parametrize("index_name", indexs)
def test_create_index(tmpdir_as_cache, index_name):
    provide = {"collection": DummyCollection()}
    index = Index.create(index_name, provide=provide)
    assert not index.exists()
    index.create_index()
    assert index.exists()


def test_anserini_get_docs(tmpdir_as_cache, dummy_index):
    docs = dummy_index.get_docs(["LA010189-0001"])
    assert docs == ["Dummy Dummy Dummy Hello world, greetings from outer space!"]
    docs = dummy_index.get_docs(["LA010189-0001", "LA010189-0002"])
    assert docs == [
        "Dummy Dummy Dummy Hello world, greetings from outer space!",
        "Dummy LessDummy Hello world, greetings from outer space!",
    ]


def test_anserini_get_df(tmpdir_as_cache, dummy_index):
    df = dummy_index.get_df("hello")
    assert df == 2


def test_anserini_get_idf(tmpdir_as_cache, dummy_index):
    idf = dummy_index.get_idf("hello")
    assert idf == 0.1823215567939546



import os
from collections import defaultdict, OrderedDict

from capreolus import ModuleBase, constants
from capreolus.utils.loginit import get_logger
from capreolus.utils.trec import topic_to_trectxt
from capreolus.utils.common import OrderedDefaultDict

logger = get_logger(__name__)  # pylint: disable=invalid-name
MAX_THREADS = constants["MAX_THREADS"]


def list2str(l, delimiter="-"):
    return delimiter.join(str(x) for x in l)


class Searcher(ModuleBase):
    """Base class for Searcher modules. The purpose of a Searcher is to query a collection via an :class:`~capreolus.index.Index` module.

    Similar to Rerankers, Searchers return a list of documents and their relevance scores for a given query.
    Searchers are unsupervised and efficient, whereas Rerankers are supervised and do not use an inverted index directly.

    Modules should provide:
        - a ``query(string)`` and a ``query_from_file(path)`` method that return document scores
    """

    module_type = "searcher"

    @staticmethod
    def load_trec_run(fn):
        # Docids in the run file appear according to decreasing score, hence it makes sense to preserve this order
        run = OrderedDefaultDict()

        with open(fn, "rt") as f:
            for line in f:
                line = line.strip()
                if len(line) > 0:
                    qid, _, docid, rank, score, desc = line.split(" ")
                    run[qid][docid] = float(score)
        return run

    @staticmethod
    def write_trec_run(preds, outfn):
        count = 0
        with open(outfn, "wt") as outf:
            qids = sorted(preds.keys(), key=lambda k: int(k))
            for qid in qids:
                rank = 1
                for docid, score in sorted(preds[qid].items(), key=lambda x: x[1], reverse=True):
                    print(f"{qid} Q0 {docid} {rank} {score} capreolus", file=outf)
                    rank += 1
                    count += 1

    def _query_from_file(self, topicsfn, output_path, cfg):
        raise NotImplementedError()

    def query_from_file(self, topicsfn, output_path):
        return self._query_from_file(topicsfn, output_path, self.config)

    def query(self, query, **kwargs):
        """
        search document based on given query, using parameters in config as default
        """
        config = {k: kwargs.get(k, self.config[k]) for k in self.config}

        cache_dir = self.get_cache_path()
        cache_dir.mkdir(exist_ok=True)
        topic_fn, runfile_dir = cache_dir / "topic.txt", cache_dir / "runfiles"

        fake_qid = "1"
        with open(topic_fn, "w", encoding="utf-8") as f:
            f.write(topic_to_trectxt(fake_qid, query))

        self._query_from_file(topic_fn, runfile_dir, config)

        runfile_fns = [f for f in os.listdir(runfile_dir) if f != "done"]
        config2runs = {}
        for runfile in runfile_fns:
            runfile_fn = runfile_dir / runfile
            runs = self.load_trec_run(runfile_fn)
            config2runs[runfile.replace("searcher_", "")] = OrderedDict(runs[fake_qid])
            os.remove(runfile_fn)  # remove it in case the file accumulate
        os.remove(runfile_dir / "done")

        return config2runs["searcher"] if len(config2runs) == 1 else config2runs


from profane import import_all_modules

from .anserini import BM25, BM25RM3, SDM

import_all_modules(__file__, __package__)



from capreolus.index import AnseriniIndex
from capreolus.tokenizer import AnseriniTokenizer


def test_anserini_tokenzier():
    cfg = {"name": "anserini", "keepstops": True, "stemmer": "none"}
    tokenizer = AnseriniTokenizer(cfg)

    index = AnseriniIndex({"name": "anserini", "indexstops": False, "stemmer": "porter", "collection": {"name": "dummy"}})
    index.create_index()

    docs = index.get_docs(["LA010189-0001"])
    print(tokenizer.tokenize(docs))



import hashlib
import os

from capreolus import ModuleBase, get_logger

logger = get_logger(__name__)


class Extractor(ModuleBase):
    """Base class for Extractor modules. The purpose of an Extractor is to convert queries and documents to a representation suitable for use with a :class:`~capreolus.reranker.Reranker` module.

    Modules should provide:
        - an ``id2vec(qid, posid, negid=None)`` method that converts the given query and document ids to an appropriate representation
    """

    module_type = "extractor"
    requires_random_seed = True

    def _extend_stoi(self, toks_list, calc_idf=False):
        if not self.stoi:
            logger.warning("extending stoi while it's not yet instantiated")
            self.stoi = {}
        # TODO is this warning working correctly?
        if calc_idf and not self.idf:
            logger.warning("extending idf while it's not yet instantiated")
            self.idf = {}
        if calc_idf and not hasattr(self, "index"):
            logger.warning("requesting calculating idf yet index is not available, set calc_idf to False")
            calc_idf = False

        n_words_before = len(self.stoi)
        for toks in toks_list:
            toks = [toks] if isinstance(toks, str) else toks
            for tok in toks:
                if tok not in self.stoi:
                    self.stoi[tok] = len(self.stoi)
                if calc_idf and tok not in self.idf:
                    self.idf[tok] = self.index.get_idf(tok)

        logger.debug(f"added {len(self.stoi)-n_words_before} terms to the stoi of extractor {self.module_name}")

    def cache_state(self, qids, docids):
        raise NotImplementedError

    def load_state(self, qids, docids):
        raise NotImplementedError

    def get_state_cache_file_path(self, qids, docids):
        """
        Returns the path to the cache file used to store the extractor state, regardless of whether it exists or not
        """
        sorted_qids = sorted(qids)
        sorted_docids = sorted(docids)
        return self.get_cache_path() / hashlib.md5(str(sorted_qids + sorted_docids).encode("utf-8")).hexdigest()

    def is_state_cached(self, qids, docids):
        """
        Returns a boolean indicating whether the state corresponding to the qids and docids passed has already
        been cached
        """
        file_path = self.get_state_cache_file_path(qids, docids)
        logger.debug("Looking for extractor cache at {}".format(file_path))
        return os.path.exists(file_path)

    def _build_vocab(self, qids, docids, topics):
        raise NotImplementedError

    def build_from_benchmark(self, *args, **kwargs):
        raise NotImplementedError

    def id2vec(self, qid, posdocid, negdocid=None, label=None):
        """
        Creates a feature from the (qid, docid) pair.
        If negdocid is supplied, that's also included in the feature (needed for training with pairwise hinge loss)
        Label is a vector of shape [num_classes], and is supplied only when using pointwise training (i.e cross entropy)
        When using pointwise samples, negdocid is None, and label is either [0, 1] or [1, 0] depending on whether the
        document represented by posdocid is relevant or irrelevant respectively.
        """
        raise NotImplementedError


from profane import import_all_modules


import_all_modules(__file__, __package__)
